unit SpiFilter;

interface

uses Windows, spi_api, NLib, ImgFilter, ifdib;

type
  TArchiveInfo = record
    FileName: string;
    FileSize: Cardinal;
    CompSize: Cardinal;
    DateTime: Int64;
    RawData: Pointer;
  end;

  TArchive = class(TObject)
  private
    FPluginName: string;
    FPluginFileName: string;
    FFileName: string;
    Handle: THandle;
    HInfo: HLOCAL;
    SpiFileInfoTop: PFileInfo;
    SpiFileInfo: PFileInfo;
    GetPluginInfo: TGetPluginInfo;
    GetArchiveInfo: TGetArchiveInfo;
    GetFile: TGetFile;
    function GetInfo(const Name: string): TArchiveInfo;
  public
    constructor Create(const PluginFileName: string);
    destructor Destroy; override;
    procedure Open(const AFileName: string);
    function EachName(var Name: string): Boolean;
    procedure Extract(const Name: string; Buf: PChar; BufSize: Longint);
    property FileName: string read FFileName;
    property Info[const Name: string]: TArchiveInfo read GetInfo; default;
    property PluginName: string read FPluginName;
    property PluginFileName: string read FPluginFileName;
  end;

  TSpiFilter = class(TImgFilter)
  private
    FFileName: string;
    GetPluginInfo: TGetPluginInfo;
    IsSupported: TIsSupported;
    GetPictureInfo: TGetPictureInfo;
    GetPicture: TGetPicture;
    ConfigurationDlg: TConfigurationDlg;
  public
    Handle: THandle;
    destructor Destroy; override;
    procedure LoadPlugin(const PluginFileName: string);
    function Test(Buf: PChar; BufSize: Cardinal): Boolean; override;
    procedure Decode(InBuf: PChar; InBufSize: Cardinal;
        var OutBuf: PChar; var OutBufSize: Cardinal; var ImgInfo: TImgInfo);
        override;
    function IsArchive: Boolean;
    function HasConfigDlg: Boolean;
    procedure ShowConfigDlg(hWnd: HWND);
    property FileName: string read FFileName;
  end;


implementation

const
  SpiErrMsg: array[0..10] of string = (
    '',
    'user cancel',
    'unknown format',
    'broken data',
    'allocate memory error',
    'memory error',
    'file read error',
    'cannot create window',
    'internal error',
    'file write error',
    'end of file error'
  );

procedure TSpiFilter.LoadPlugin(const PluginFileName: string);
var
  buf: array[0..255] of Char;
begin
  Handle := LoadLibrary(PChar(PluginFileName));
  if Handle <> 0 then
  begin
    FFileName := PluginFileName;
    GetPluginInfo := GetProcAddress(Handle, 'GetPluginInfo');
    IsSupported := GetProcAddress(Handle, 'IsSupported');
    GetPictureInfo := GetProcAddress(Handle, 'GetPictureInfo');
    GetPicture := GetProcAddress(Handle, 'GetPicture');
    ConfigurationDlg := GetProcAddress(Handle, 'ConfigurationDlg');
    if Assigned(GetPluginInfo) then
    begin
      GetPluginInfo(1, buf, SizeOf(buf));
      FName := buf;
    end;
  end;
end;

destructor TSpiFilter.Destroy;
begin
  if Handle <> 0 then
    FreeLibrary(Handle);
end;

function TSpiFilter.Test(Buf: PChar; BufSize: Cardinal): Boolean;
var
  tmpbuf: array[0..2047] of Char;
begin
  // Call IsSupported() with copyed buffer. IsSupprted() may destroy buffer.
  if BufSize > SizeOf(tmpbuf) then
    BufSize := SizeOf(tmpbuf);
  FillChar(tmpbuf, SizeOf(tmpbuf), 0);
  Move(Buf^, tmpbuf, BufSize);
  Result := Assigned(IsSupported) and (IsSupported('', DWORD(@tmpbuf)) <> 0);
end;

function ProgressCallback(nNum, nDemon: Integer; lData : Longint): Integer; stdcall;
begin
  Result := 0;
end;

procedure SpiCheck(Code: Integer; const Name: string);
var
  msg: string;
begin
  if Code = 0 then
    Exit;
  if Code <= High(SpiErrMsg) then
    msg := SpiErrMsg[Code] + ' ('+IStr(Code)+')'
  else
    msg := 'error ' + IStr(Code);
  raise EImgFilterError.Create(Name+': '+msg);
end;

procedure TSpiFilter.Decode(InBuf: PChar; InBufSize: Cardinal;
    var OutBuf: PChar; var OutBufSize: Cardinal; var ImgInfo: TImgInfo);
var
  hbinfo, hbm: HLOCAL;
  size: Integer;
  bmi: TMaxBitmapInfo;
  picinfo: TPictureInfo;
begin
  if not Assigned(GetPictureInfo) then
    raise EImgFilterError.Create('GetPictureInfo not found');

  // For plugin which do not set all picinfo paramateres.
  FillChar(picinfo, SizeOf(picinfo), 0);
  SpiCheck(GetPictureInfo(InBuf, InBufSize, 1, picinfo), 'GetPictureInfo');

  with ImgInfo do
  begin
    Width := picinfo.width;
    Height := picinfo.height;
    Bpp := picinfo.colorDepth;
    FormatName := '';
    if picinfo.hInfo = 0 then
      Comment := ''
    else begin
      SetString(Comment, PChar(LocalLock(picinfo.hInfo)),
        LocalSize(picinfo.hInfo)-1); // -1: for last null character
      LocalUnLock(picinfo.hInfo);
      LocalFree(picinfo.hInfo);
    end;
    Components := 3; // DIB decoded by SPI is always 24bpp(3bytes).
  end;

  if not Assigned(GetPicture) then
    raise EImgFilterError.Create('GetPicture not found');

  // Call GetPicture() with dummy ProgressCallback function.
  // Some plugin may cause error if ProgressCallback is nil.
  SpiCheck(GetPicture(InBuf, InBufSize, 1, hbinfo, hbm, ProgressCallback, 0),
      'GetPicture');

  size := LocalSize(hbinfo);
  if size > SizeOf(TMaxBitmapInfo) then
    size := SizeOf(TMaxBitmapInfo);
  Move(PMaxBitmapInfo(LocalLock(hbinfo))^, bmi, size);
  LocalUnLock(hbinfo);
  LocalFree(hbinfo);

  OutBufSize := ((ImgInfo.Width * 3 + 3) and not 3) * ImgInfo.Height;
  OutBuf := Alloc(OutBufSize);
  if OutBuf = nil then
    raise EImgFilterError.Create('output buffer allocation failed');

  DecodeDIB(LocalLock(hbm), LocalSize(hbm), bmi, OutBuf, OutBufSize);
  LocalUnLock(hbm);
  LocalFree(hbm);
end;

function TSpiFilter.IsArchive: Boolean;
var
  ver: array[0..15] of Char;
begin
  FillChar(ver, SizeOf(ver), 0);
  if Assigned(GetPluginInfo) then
    GetPluginInfo(0, ver, SizeOf(ver));
  Result := Copy(ver, 1, 4) = '00AM';
end;

function TSpiFilter.HasConfigDlg: Boolean;
begin
  Result := Assigned(ConfigurationDlg);
end;

procedure TSpiFilter.ShowConfigDlg(hWnd: HWND);
begin
  SpiCheck(ConfigurationDlg(hWnd, 1), 'ConfigurationDlg');
end;


{ TArchive }

constructor TArchive.Create(const PluginFileName: string);
var
  buf: array[0..255] of Char;
begin
  Handle := LoadLibrary(PChar(PluginFileName));
  if Handle <> 0 then
  begin
    FPluginFilename := PluginFileName;
    GetArchiveInfo := GetProcAddress(Handle, 'GetArchiveInfo');
    GetFile := GetProcAddress(Handle, 'GetFile');
    GetPluginInfo := GetProcAddress(Handle, 'GetPluginInfo');
    if Assigned(GetPluginInfo) then
    begin
      GetPluginInfo(1, buf, SizeOf(buf));
      FPluginName := buf;
    end;
  end;
end;

procedure TArchive.Open(const AFileName: string);
begin
  FFileName := AFileName;
  if not Assigned(GetArchiveInfo) then
    raise EImgFilterError.Create('GetArchiveInfo not found');
  SpiCheck(GetArchiveInfo(PChar(FFileName), 0, 0, HInfo), 'GetArchiveInfo');
  SpiFileInfoTop := LocalLock(HInfo);
  SpiFileInfo := SpiFileInfoTop;
end;

destructor TArchive.Destroy;
begin
  if HInfo <> 0 then
  begin
    LocalUnlock(HInfo);
    LocalFree(HInfo);
  end;
  if Handle <> 0 then
    FreeLibrary(Handle);
end;

function TArchive.GetInfo(const Name: string): TArchiveInfo;
var
  spiinfo: PFileInfo;
begin
  spiinfo := SpiFileInfoTop;
  while spiinfo.method[0] <> #0 do
    if Name = spiinfo.path + spiinfo.filename then
    begin
      Result.FileName := Name;
      Result.FileSize := spiinfo.filesize;
      Result.CompSize := spiinfo.compsize;
      Result.DateTime := spiinfo.timestamp;
      Result.RawData := spiinfo;
      Exit;
    end else
      Inc(spiinfo);
  Result.FileName := '';
  Result.RawData := nil;
end;

function TArchive.EachName(var Name: string): Boolean;
begin
  Result := SpiFileInfo.method[0] <> #0;
  if Result then
    Name := SpiFileInfo.path + SpiFileInfo.filename;
  Inc(SpiFileInfo);
end;

procedure TArchive.Extract(const Name: string; Buf: PChar; BufSize: Longint);
var
  info: PFileInfo;
  hdst: HLOCAL;
  p: PChar;
  fsize: Integer;
begin
  info := GetInfo(Name).RawData;
  if info = nil then
    raise EImgFilterError.Create('Archive file info not found');

  if not Assigned(GetFile) then
    raise EImgFilterError.Create('GetFile not found');
  SpiCheck(
      GetFile(PChar(FFileName), info.position, PChar(@hdst), $100, nil,0),
      'GetFile');
  p := LocalLock(hdst);
  fsize := LocalSize(hdst);
  if fsize > BufSize then
    fsize := BufSize;
  Move(p^, Buf^, fsize);
  LocalUnLock(hdst);
  LocalFree(hdst);
end;

end.
